import pyinotify


class WatchableFolder(pyinotify.ProcessEvent):

    def __init__(self, path='', recursive=False, listen_to=None):
        self.path = path
        self.recursive = recursive
        self.listen_to = listen_to

    def process_IN_ACCESS(self, event):
        """A file was accessed."""

    def process_IN_ATTRIB(self, event):
        """Metadata changed for a file."""

    def process_IN_CLOSE_NOWRITE(self, event):
        """An unwritable file was closed."""

    def process_IN_CLOSE_WRITE(self, event):
        """A writable file was closed."""

    def process_IN_CREATE(self, event):
        """A file/directory was created in watched directory."""

    def process_IN_DELETE(self, event):
        """A file/directory was deleted in watched directory."""

    def process_IN_DELETE_SELF(self, event):
        """The watched item itself was deleted."""

    def process_IN_ISDIR(self, event):
        """Any event occurred that was on a directory."""

    def process_IN_MODIFY(self, event):
        """A file was modified."""

    def process_IN_MOVE_SELF(self, event):
        """The watched item itself was moved somewhere."""

    def process_IN_MOVED_FROM(self, event):
        """A file/directory was moved away from the current watched directory."""

    def process_IN_MOVED_TO(self, event):
        """A file/directory was moved into the current watched directory."""

    def process_IN_OPEN(self, event):
        """A file was opened."""

    def process_IN_UNMOUNT(self, event):
        """The file system the watched directory is associated with was unmounted."""



"""
Thanks to Tim Golden for his write-up on events with pywin32:
http://timgolden.me.uk/python/win32_how_do_i/watch_directory_for_changes.html
"""

import os
from threading import Thread

import win32con
import win32file


class FileEvent(object):

    def __init__(self, file_type, name, full_path, maskname):
        self.dir = True if file_type == 'dir' else False
        self.name = name
        self.pathname = full_path
        self.maskname = maskname


class FolderWatcher(Thread):

    ACTIONS = {
        1: "IN_CREATE",
        2: "IN_DELETE",
        3: "IN_MODIFY",
        4: "IN_RENAMED_FROM",
        5: "IN_RENAMED_TO"
    }

    FILE_LIST_DIRECTORY = 0x0001

    def __init__(self, folder, queue, **kwargs):
        super(FolderWatcher, self).__init__(**kwargs)

        self.setDaemon(1)
        self.folder = folder
        self.queue = queue
        self.start()

    def run(self):
        for events in self._watch():
            self.queue.put((self.folder, events))

    def _watch(self):
        w_dir = win32file.CreateFile(
            self.folder.path,
            self.FILE_LIST_DIRECTORY,
            win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,
            None,
            win32con.OPEN_EXISTING,
            win32con.FILE_FLAG_BACKUP_SEMANTICS,
            None
        )

        while 1:
            results = win32file.ReadDirectoryChangesW(
                w_dir,
                1024,
                self.folder.recursive,
                win32con.FILE_NOTIFY_CHANGE_FILE_NAME |
                win32con.FILE_NOTIFY_CHANGE_DIR_NAME |
                win32con.FILE_NOTIFY_CHANGE_ATTRIBUTES |
                win32con.FILE_NOTIFY_CHANGE_SIZE |
                win32con.FILE_NOTIFY_CHANGE_LAST_WRITE |
                win32con.FILE_NOTIFY_CHANGE_SECURITY,
                None,
                None
            )

            for action, name in results:
                if action not in self.ACTIONS:
                    continue

                maskname = self.ACTIONS[action]
                if not hasattr(self.folder, 'process_{0}'.format(maskname)):
                    continue

                full_path = os.path.join(self.folder.path, name)
                if not os.path.exists(full_path):
                    file_type = 'deleted'
                elif os.path.isdir(full_path):
                    file_type = 'dir'
                else:
                    file_type = 'file'

                events = [FileEvent(file_type, name, full_path, maskname)]

                if file_type == 'dir':
                    events.append(FileEvent(file_type, name, full_path, 'IN_ISDIR'))

                yield events



import platform

if platform.system() == 'Windows':
    from .windows import WatchableFolder, FolderSpy
elif platform.system() == 'Linux':
    from .linux import WatchableFolder, FolderSpy
else:
    raise Exception('{0} is not a supported.'.format(platform.system()))



import queue

from .watcher import FolderWatcher
from folderspy.spy import Spy


class FolderSpy(Spy):

    @staticmethod
    def _event_loop(folders):
        """Kicks off the event loop that begins watching for events."""

        event_queue = queue.Queue()
        for folder in folders:
            FolderWatcher(folder, queue=event_queue)

        while 1:
            try:
                folder, events = event_queue.get_nowait()
                for event in events:
                    getattr(folder, 'process_{0}'.format(event.maskname))(event)
            except queue.Empty:
                pass



import sys
import logging
from abc import abstractmethod


class Spy(object):

    @classmethod
    def watch(cls, *args):
        """Set up watches and begin loop to trigger events on folders."""

        cls._setup_watchers(args)
        try:
            cls._event_loop(args)
        except KeyboardInterrupt:
            sys.exit(0)

    @staticmethod
    def _event_loop(folders):
        """Kicks off the event loop that begins watching for events."""

    @staticmethod
    def _setup_watchers(folders):
        """Setup watchers for all folders."""



import asyncore

import pyinotify

from folderspy.spy import Spy


class FolderSpy(Spy):

    @staticmethod
    def _setup_watchers(folders):
        """Setup watchers for all folders."""

        manager = pyinotify.WatchManager()
        pyinotify.AsyncNotifier(manager)

        for folder in folders:
            manager.add_watch(
                folder.path,
                mask=folder.listen_to,
                proc_fun=folder,
                rec=folder.recursive
            )

    @staticmethod
    def _event_loop(folders):
        """Kicks off the event loop that begins watching for events."""

        asyncore.loop()



class WatchableFolder(object):

    def __init__(self, path='', recursive=False, listen_to=None):
        self.path = path
        self.recursive = recursive
        self.listen_to = listen_to

    def process_IN_CREATE(self, event):
        """A file/directory was created in watched directory."""

    def process_IN_DELETE(self, event):
        """A file/directory was deleted in watched directory."""

    def process_IN_MODIFY(self, event):
        """A file was modified."""

    def process_IN_RENAMED_FROM(self, event):
        """The name the file/directory was renamed from."""

    def process_IN_RENAMED_TO(self, event):
        """The name the file/directory was renamed to."""

    def process_IN_ISDIR(self, event):
        """Any event occurred that was on a directory."""



"""A setuptools based setup module.

https://github.com/requests/requests/blob/master/setup.py
"""
import os
import sys
from codecs import open
from shutil import rmtree

from setuptools import setup, find_packages, Command

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

about = {}

with open(os.path.join(here, 'folderspy', '__version__.py')) as f:
    exec(f.read(), about)


class PublishCommand(Command):
    """Support setup.py publish."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def clean_up_builds():
        try:
            print('Removing any previous builds')
            rmtree(os.path.join(here, 'dist'))
            rmtree(os.path.join(here, 'build'))
            rmtree([os.path.join(here, dir) for dir in os.listdir(here) if dir.endswith('.egg-info')][0])
        except FileNotFoundError:
            pass

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.clean_up_builds()

        print('Building Source and Wheel (universal) distribution')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        print('Uploading the package to PyPI via Twine...')
        os.system('twine upload dist/*')

        print('Pushing git tags...')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()

setup(
    name='folderspy',
    version=about['__version__'],
    description='Watch folders for file/directory events with a simple API.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/rob-white/folderspy',
    author='Rob White',
    author_email='robathan4@gmail.com',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'pyinotify ; platform_system == "Linux"',
        'pypiwin32 ; platform_system == "Windows"'
    ],
    cmdclass={
        'publish': PublishCommand
    }
)
